#include <windows.h>
#include <io.h>
#include <memory.h>
#include <fcntl.h>
#include <dinput.h>

#include "type.h"
#include "joydrv.h"
#include "win32def.h"

#define RETURN { JOY1_Close(); return FPSE_ERR; }

#define DIRELEASE(d) if (d != NULL) { d->Release(); d = NULL; }

#define DINPUT_BUFFERSIZE  16

DIPROPDWORD dipdw = {
// the header
        {
            sizeof(DIPROPDWORD),        // diph.dwSize
            sizeof(DIPROPHEADER),       // diph.dwHeaderSize
            0,                          // diph.dwObj
            DIPH_DEVICE,                // diph.dwHow
        },
// the data
        DINPUT_BUFFERSIZE,              // dwData
};

// DirectInput vars
static LPDIRECTINPUT       lpdi      = NULL;
static LPDIRECTINPUTDEVICE lpdiMouse = NULL;

static HANDLE  g_hevtMouse;
static char   *padbuf = NULL;

//-----------------------------------------------------------------

// This Function is called when emulator starts
int JOY1_Open(UINT32 *par)
{
    FPSEWin32     *joy = (FPSEWin32 *)par;
    HRESULT        hr; 

    hr = DirectInputCreate(joy->HInstance,
							DIRECTINPUT_VERSION, &lpdi, NULL); 
    if (hr != DI_OK) RETURN
    hr = lpdi->CreateDevice(GUID_SysMouse, &lpdiMouse, NULL);
    if (FAILED(hr)) RETURN
    hr = lpdiMouse->SetDataFormat(&c_dfDIMouse);
    if (FAILED(hr)) RETURN
    hr = lpdiMouse->SetCooperativeLevel(joy->HWnd,
               DISCL_EXCLUSIVE | DISCL_FOREGROUND);
    if (FAILED(hr)) RETURN

    g_hevtMouse = CreateEvent(0, 0, 0, 0);
    if (g_hevtMouse == NULL) RETURN;
    hr = lpdiMouse->SetEventNotification(g_hevtMouse);
    if (FAILED(hr)) RETURN

    hr = lpdiMouse->SetProperty(DIPROP_BUFFERSIZE, &dipdw.diph);
    if (FAILED(hr)) RETURN

    hr = lpdiMouse->Acquire();
    if (FAILED(hr)) RETURN

    return FPSE_OK;
}

// This Function is called when emulator is closed
void JOY1_Close()
{
    DIRELEASE(lpdiMouse);
    DIRELEASE(lpdi);
}

// This function is called when the gamepad should vibrate:
// if (motor != 0) then vibrate(on) else vibrate(off)
void JOY1_Vibrate(int motor)
{
}

// Return status of buttons (see joydrv.h)
void JOY1_Poll()
{
    DIMOUSESTATE    mouse;
    HRESULT         res;
    int             x;

    do
    {
        res = lpdiMouse->GetDeviceState(sizeof(mouse), &mouse);
        if (res == DIERR_INPUTLOST ||
            res == DIERR_NOTACQUIRED) lpdiMouse->Acquire();
    } while (res == DIERR_INPUTLOST);

    if (res != DI_OK)
    {
        padbuf[4] = 0xFF;
        padbuf[5] = 0;
        padbuf[6] = 0;
        return;
    }

    x = mouse.lX;
    if (x < -128) x = -128;
        else
    if (x > 127) x = 127;
    padbuf[5] = x;

    x = mouse.lY;
    if (x < -128) x = -128;
        else
    if (x > 127) x = 127;
    padbuf[6] = x;

    x = 0xFF;
    if (mouse.rgbButtons[0] & 0x80) x &= ~0x08;
    if (mouse.rgbButtons[1] & 0x80) x &= ~0x04;
    padbuf[4] = x;
}

int JOY1_InitHeader(char *buf)
{
    padbuf = buf;
    buf[1] = 0x12;
    buf[2] = 0x5A;
    buf[3] = 0xFF;
    buf[4] = 0xFF;
    buf[5] = 0;
    buf[6] = 0;
    return 7;
}

int JOY1_Configure(UINT32 *par)
{
    return FPSE_OK;
}

void JOY1_About(UINT32 *par)
{
}
