#define INITGUID
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <ddraw.h>
#include <d3d.h>

#include "gpulocal.h"
#include "eb_macro.h"

#define GPUMAXEB    1024
#define FORCE_SOFT	0

#define CLIPPER(sx,mx,sw,cx1,cx2,t)	\
{					\
	int v;				\
					\
	if (cx1 >= cx2) return;		\
	v = cx1 - sx;			\
	if (v>0) {			\
		if ( sw < v ) return;	\
		sw -= v;		\
		t += v;			\
		sx = cx1;		\
	}				\
	if ( sx > cx2 ) return ;        \
	v=cx2-sx;			\
	if ( sw >= v ) sw = v - 1;	\
}

// import clipping & offset
#define XMIN    gpu.clip_min.x
#define XMAX    gpu.clip_max.x
#define YMIN    gpu.clip_min.y
#define YMAX    gpu.clip_max.y
#define OFFX    gpu.drawoffset.x
#define OFFY    gpu.drawoffset.y

// PSX(24/16) to VGA(16) RGB conversion
#define COLORMODEL	555 // Change to '555' or '565'

#if COLORMODEL==555
#define RGB32_16(c) (((c) & 0x00F80000) >> 9) | \
                    (((c) & 0x0000F800) >> 6) | \
                    (((c) & 0x000000F8) >> 3)
#define PSX2VGA(c)	(((c) & 0x001F) << 10) | \
					(((c) & 0x83E0)      ) | \
					(((c) & 0x7F00) >> 10)
#else
#define RGB32_16(c) (((c) & 0x00F80000) >> 8) | \
                    (((c) & 0x0000FC00) >> 5) | \
                    (((c) & 0x000000F8) >> 3)
#endif

#define FILL_TRIANGLE(tri_flags,ptr)                       \
		{    ((LPD3DTRIANGLE)ptr)->wV1    = order[0];      \
			 ((LPD3DTRIANGLE)ptr)->wV2    = order[1];      \
			 ((LPD3DTRIANGLE)ptr)->wV3    = order[2];      \
			 ((LPD3DTRIANGLE)ptr)->wFlags = tri_flags;     \
		     ptr = (void *)(((LPD3DTRIANGLE)(ptr)) + 1); }

// DirectDraw definitions
#define DDCLEAR(dd)     memset(&dd,0,sizeof(dd)); dd.dwSize = sizeof(dd);
#define RELEASE(dd)     if (dd != NULL) { dd->lpVtbl->Release(dd); dd = NULL; }

// User-defined types
typedef struct {
	LPDIRECTDRAWSURFACE	surface;
	D3DTEXTUREHANDLE	handle;
	UINT32				tpage,clutid;
} D3DCache;

// DirectDraw interfaces 
 
static LPDIRECTDRAW            lpdd                      = NULL; 
static LPDIRECTDRAWSURFACE     lpddPrimary               = NULL; 
static LPDIRECTDRAWSURFACE     lpddDevice                = NULL; 
 
// Direct3D interfaces 
 
static LPDIRECT3D              lpd3d                     = NULL; 
static LPDIRECT3DDEVICE        lpd3dDevice               = NULL; 
static LPDIRECT3DVIEWPORT      lpd3dViewport             = NULL; 
static LPDIRECT3DEXECUTEBUFFER lpd3dExecuteBuffer        = NULL; 

static int                     FullScreenMode=0;

// Globals used for selecting the Direct3D device. They are 
// globals because this makes it easy for the enumeration callback 
// function to read and write from them. 
 
static DWORD                   dwDeviceBitDepth          = 0; 
 
static void DrawBoxNoClip(int x,int y,int w,int h,int c);

static unsigned char overtbl[256]; /* 0..127..255 -> 0..255..255 */
static unsigned short rgbtbl[65536]; /* bgr -> rgb */

#define	CACHE_MAX	16

static D3DCache tcache[CACHE_MAX];
static int      tcache_index;

static void DrawInit(void)
{
	int i,r,g,b;
	for(i=0;i<128;i++) overtbl[i]=i*2;
	for(;i<256;i++) overtbl[i]=255;
	i = 0;
	for(b=0;b<32;b++)
		for(g=0;g<32;g++)
			for(r=0;r<32;r++,i++)
				rgbtbl[i] = (r<<10) | (g<<5) | b;
	for(i=0;i<32768;i++)
		rgbtbl[i+32768]=rgbtbl[i]; // | 0x8000;
	memset(tcache,0,CACHE_MAX*sizeof(D3DCache));
}

void win_resize(int w,int h)
{
        RECT window,client;
        int ew,eh;

		if (!FullScreenMode)
		{
        GetWindowRect(hwnd,&window);
        GetClientRect(hwnd,&client);
        ew = (window.right-window.left)-(client.right-client.left),
        eh = (window.bottom-window.top)-(client.bottom-client.top),

        MoveWindow(hwnd,window.left,window.top,w+ew,h+eh,TRUE);
		}
}

static LPDIRECTDRAWSURFACE DD_alloc(int w,int h,int flags)
{
    HRESULT             hRet;
    DDSURFACEDESC       ddsd;
    DDCOLORKEY          key;
    LPDIRECTDRAWSURFACE DDs;

    DDCLEAR(ddsd);
    memset(&key,0,sizeof(DDCOLORKEY));
    DDs = NULL;
    ddsd.dwFlags  = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = flags; // | DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth  = w;
    ddsd.dwHeight = h;

    hRet = IDirectDraw_CreateSurface(lpdd,&ddsd, &DDs, NULL);
    if (hRet != DD_OK) return NULL;

    hRet = IDirectDrawSurface_SetColorKey(DDs, DDCKEY_SRCBLT, &key);
    if (hRet != DD_OK) return NULL;

    return DDs;
}

static D3DTEXTUREHANDLE D3D_alloc(LPDIRECTDRAWSURFACE surface)
{
	LPDIRECT3DTEXTURE texture;
	D3DTEXTUREHANDLE  handle;

	IDirect3D_QueryInterface(surface,&IID_IDirect3DTexture,&texture);
	IDirect3DTexture_GetHandle(texture,lpd3dDevice,&handle);
	RELEASE(texture);

	return handle;
}

static void pixel2texel(UINT16 *t, UINT16 *p,int n)
{
	do {
		*t++=rgbtbl[*p++];
	} while(--n);
}

//static int pos = 0;
static void load_texture(D3DCache *texture,int tpage,int clut)
{
	HDC hdc;
	UINT16 *ctbl,*tex;
	UINT16 *t,ctbl2[256],texturebuf[256*256];
	int x,y,mode;

	struct {
		BITMAPINFOHEADER	bmiHeader;
		RGBQUAD	bmiColors[256];
	} bi;
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth  =  256;
	bi.bmiHeader.biHeight = -256;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 16;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = 0;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;

	mode = (tpage>>7)&3;

	ctbl = vram+(clut&0x7fff)*16;
	tex  = vram+(tpage&15)*64 + (tpage&16)*16*FRAME_W;
	t = texturebuf;

	switch (mode) {
	case 0: /* 4bit */
		pixel2texel(ctbl2,ctbl,16);
		for(y=0;y<256;y++) {
			for(x=0;x<256;x+=4) {
				int c = *tex++;
				t[0] = ctbl2[c&15];
				t[1] = ctbl2[(c>>4)&15];
				t[2] = ctbl2[(c>>8)&15];
				t[3] = ctbl2[(c>>12)];
				t+=4;
			}
			tex += FRAME_W-256/4;
		}
		break;

	case 1: /* 8bit */
		pixel2texel(ctbl2,ctbl,256);
		for(y=0;y<256;y++) {
			for(x=0;x<256;x+=2) {
				int c = *tex++;
				t[0] = ctbl2[c&255];
				t[1] = ctbl2[c>>8];
				t+=2;
			}
			tex += FRAME_W-256/2;
		}
		break;

	case 2: /* 16bit */
		for(y=0;y<256;y++) {
			pixel2texel(t,tex,256);
			tex += FRAME_W;
			t += 256;
		}
		break;
	}

	IDirectDrawSurface_GetDC(texture->surface,&hdc);
	StretchDIBits(hdc,
		0,0,256,256, 
		0,0,256,256, 
	texturebuf,(BITMAPINFO*)&bi,DIB_RGB_COLORS,SRCCOPY);

	IDirectDrawSurface_ReleaseDC(texture->surface,hdc);
/*
	IDirectDrawSurface_GetDC(lpddDevice,&hdc);
	StretchDIBits(hdc,
		pos*256,0,256,256, 
		0,0,256,256, 
	texturebuf,(BITMAPINFO*)&bi,DIB_RGB_COLORS,SRCCOPY);
pos = (pos+1)&1;
	IDirectDrawSurface_ReleaseDC(lpddDevice,hdc);
*/
}

static D3DCache *get_texture(UINT32 tpage, UINT32 clut)
{
	D3DCache *tc;
	int       tchanged,cchanged,i;

	tpage&=(3<<7)|31;
	clut &= 0x7fff;

	tchanged = tpage_changed(tpage);
	cchanged = clut_changed(clut);

	if (tchanged | cchanged) {
		/* parge cache */
		for(i=0, tc = tcache;i<CACHE_MAX;i++, tc++) {
			if ((cchanged && tc->clutid==clut )
			||  (tchanged && tc->tpage==tpage)) {
				tc->clutid = -1;
				tc->tpage  = -1;
			}
		}
	} else {
		/* search cache */
		for(i=0, tc = tcache;i<CACHE_MAX;i++, tc++) {
			if (tc->tpage==tpage && tc->clutid==clut)
				return tc;
		}
	}
	printf("Texture Cache miss hit tpage=%04x clut=%04x\n",tpage,clut);
	tc = &tcache[tcache_index];
	if (++tcache_index >= CACHE_MAX) tcache_index = 0;

	tc->clutid = clut;
	tc->tpage  = tpage;

	if (tc->surface == NULL)
	{
		tc->surface = DD_alloc(256,256,DDSCAPS_TEXTURE);
		tc->handle  = D3D_alloc(tc->surface);
	}
	load_texture(tc,tpage,clut);

	return tc;
}

static HRESULT CreatePrimary() 
{ 
    HRESULT             hRes; 
    DDSURFACEDESC       ddsd; 
    LPDIRECTDRAWCLIPPER lpddClipper; 
 
    DDCLEAR(ddsd);
    ddsd.dwFlags        = DDSD_CAPS; 
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE; 
    hRes = lpdd->lpVtbl->CreateSurface(lpdd, &ddsd, &lpddPrimary, NULL); 
    if (FAILED(hRes)) 
        return hRes; 
 
    hRes = DirectDrawCreateClipper(0, &lpddClipper, NULL); 
    if (FAILED(hRes)) 
        return hRes; 
    hRes = lpddClipper->lpVtbl->SetHWnd(lpddClipper, 0, hwnd); 
    if (FAILED(hRes)) 
    { 
        lpddClipper->lpVtbl->Release(lpddClipper); 
        return hRes; 
    } 
    hRes = lpddPrimary->lpVtbl->SetClipper(lpddPrimary, lpddClipper); 
    if (FAILED(hRes)) 
    { 
        lpddClipper->lpVtbl->Release(lpddClipper); 
        return hRes; 
    } 
 
    lpddClipper->lpVtbl->Release(lpddClipper); 
 
    return DD_OK; 
} 
 
static HRESULT CreateDirect3D() 
{ 
    HRESULT hRes; 
 
    hRes = DirectDrawCreate(NULL, &lpdd, NULL); 
    if (FAILED(hRes)) 
        return hRes; 
 
    hRes = lpdd->lpVtbl->SetCooperativeLevel(lpdd, hwnd, DDSCL_NORMAL); 
    if (FAILED(hRes)) 
        return hRes; 
 
    hRes = lpdd->lpVtbl->QueryInterface(lpdd, &IID_IDirect3D, &lpd3d); 
    if (FAILED(hRes)) 
        return hRes; 
 
    return DD_OK; 
} 
 
static HRESULT CreateDevice(DWORD dwWidth, DWORD dwHeight) 
{ 
    HRESULT         hRes; 

	lpddDevice = DD_alloc(1024,512,DDSCAPS_3DDEVICE);
	if (lpddDevice == NULL) return E_FAIL;

#if !FORCE_SOFT
    hRes = lpddDevice->lpVtbl->QueryInterface(lpddDevice, 
                                   &IID_IDirect3DHALDevice,
								   &lpd3dDevice); 
    if (FAILED(hRes))
	{
#endif
	    hRes = lpddDevice->lpVtbl->QueryInterface(lpddDevice, 
                                   &IID_IDirect3DRGBDevice,
								   &lpd3dDevice);
		if (FAILED(hRes))
			return hRes; 
		printf("Using RGB HEL render\n");
#if !FORCE_SOFT
	} else printf("Using HAL render\n");
#endif

    return DD_OK; 
} 

static HRESULT CreateScene(void) 
{ 
    HRESULT              hRes; 
    D3DEXECUTEBUFFERDESC d3dExecuteBufferDesc; 
//    D3DEXECUTEDATA       d3dExecuteData; 
 
    // Create the viewport. 
    // The viewport parameters are set in the UpdateViewport function, 
    // which is called in response to WM_SIZE. 
 
    hRes = lpd3d->lpVtbl->CreateViewport(lpd3d, &lpd3dViewport, NULL); 
    if (FAILED(hRes)) 
        return hRes; 
    hRes = lpd3dDevice->lpVtbl->AddViewport(lpd3dDevice, lpd3dViewport); 
    if (FAILED(hRes)) 
        return hRes; 
 
    // Build the execute buffer. 

    DDCLEAR(d3dExecuteBufferDesc); 
    d3dExecuteBufferDesc.dwFlags      = D3DDEB_BUFSIZE; 
    d3dExecuteBufferDesc.dwBufferSize = GPUMAXEB; 
    hRes = lpd3dDevice->lpVtbl->CreateExecuteBuffer(lpd3dDevice, 
        &d3dExecuteBufferDesc, &lpd3dExecuteBuffer, NULL); 
    if (FAILED(hRes)) return hRes; 

    return DD_OK; 
} 
 
static HRESULT RenderScene(char *buf, int nvert, int istr_size) 
{ 
    HRESULT hRes; 
    D3DEXECUTEBUFFERDESC d3dExeBufDesc; 
    D3DEXECUTEDATA       d3dExecuteData; 

    DDCLEAR(d3dExeBufDesc);
    DDCLEAR(d3dExecuteData);
    d3dExecuteData.dwVertexCount       = nvert; 
    d3dExecuteData.dwInstructionOffset = nvert*sizeof(D3DTLVERTEX);
    d3dExecuteData.dwInstructionLength = istr_size;
    hRes = lpd3dExecuteBuffer->lpVtbl->SetExecuteData 
        (lpd3dExecuteBuffer, &d3dExecuteData); 
    if (FAILED(hRes)) return hRes; 

    hRes = lpd3dExecuteBuffer->lpVtbl->Lock(lpd3dExecuteBuffer, 
        &d3dExeBufDesc); 
    if (FAILED(hRes)) return hRes;

	memcpy(d3dExeBufDesc.lpData,buf,nvert*sizeof(D3DTLVERTEX)+istr_size);

    lpd3dExecuteBuffer->lpVtbl->Unlock(lpd3dExecuteBuffer); 

// Start the scene. 
    hRes = lpd3dDevice->lpVtbl->BeginScene(lpd3dDevice); 
    if (FAILED(hRes)) return hRes; 
 
    hRes = lpd3dDevice->lpVtbl->Execute(lpd3dDevice, lpd3dExecuteBuffer, 
        lpd3dViewport, D3DEXECUTE_UNCLIPPED); 
    if (FAILED(hRes)) 
    { 
        lpd3dDevice->lpVtbl->EndScene(lpd3dDevice); 
        return hRes; 
    } 
 
// End the scene. 
    hRes = lpd3dDevice->lpVtbl->EndScene(lpd3dDevice); 
    if (FAILED(hRes)) 
        return hRes; 
 
    return DD_OK; 
} 
 
static HRESULT UpdateViewport(void) 
{ 
    D3DVIEWPORT d3dViewport; 
 
    ZeroMemory(&d3dViewport, sizeof(d3dViewport)); 
    d3dViewport.dwSize   = sizeof(d3dViewport); 
    d3dViewport.dwX      = 0; 
    d3dViewport.dwY      = 0; 
    d3dViewport.dwWidth  = (DWORD)639; //rSrcRect.right; 
    d3dViewport.dwHeight = (DWORD)511; //rSrcRect.bottom; 
    d3dViewport.dvScaleX = D3DVAL((float)d3dViewport.dwWidth/2); 
    d3dViewport.dvScaleY = D3DVAL((float)d3dViewport.dwHeight/2); 
    d3dViewport.dvMaxX   = D3DVAL(1.0); 
    d3dViewport.dvMaxY   = D3DVAL(1.0); 
    return lpd3dViewport->lpVtbl->SetViewport(lpd3dViewport, 
        &d3dViewport); 
} 
 
int D3D_Open(void)
{
//	HRESULT hRes;

	DrawInit();
    if (FAILED(CreateDirect3D())) 
    { 
		D3D_Close(); 
        return FPSE_ERR; 
    } 
 
    if (FAILED(CreatePrimary())) 
    {
		D3D_Close();
        return FPSE_ERR; 
    } 

	CreateDevice(640,512);
	CreateScene();
	UpdateViewport();
	
	DrawBoxNoClip(0,0,640,512,0);
    return FPSE_OK;
}

void D3D_Close(void)
{
	int x;

	for (x=0; x<CACHE_MAX; x++)
		RELEASE(tcache[x].surface);

    RELEASE(lpd3dExecuteBuffer);
    RELEASE(lpd3dViewport);
    RELEASE(lpd3dDevice);
    RELEASE(lpddDevice);
    RELEASE(lpddPrimary);
	RELEASE(lpd3d);
	RELEASE(lpdd);
}

void D3D_update(void)
{
	RECT r,r2;

	if (lpdd == NULL) return;

	r2.left = r2.top = 0;

	r.left   = gpu.disp.x;
	r2.right = gpu.disp.w;
	r.right  = gpu.disp.x + gpu.disp.w;
	r.top    = gpu.disp.y;
	r2.bottom = gpu.disp.h;
	r.bottom = gpu.disp.y + gpu.disp.h;

	switch (FullScreenMode) {
	case 0:
		ClientToScreen(hwnd,(LPPOINT)(&(r2.left)));
		ClientToScreen(hwnd,(LPPOINT)(&(r2.right)));
	case 1:
		break;
	case 2:
		r2.right = 639;
		r2.bottom = 479;
		break;
	}

    IDirectDrawSurface_Blt(lpddPrimary,
						   &r2,
						   lpddDevice,
						   &r,
						   DDBLT_WAIT,
						   NULL);
}

static void DrawBoxNoClip(int x,int y,int w,int h,int c)
{
	    DDBLTFX efx;
        RECT    r;

        r.left   = x;
        r.right  = x+w;
        r.top    = y;
        r.bottom = y+h;

        if (r.left > 1023) return;
        if (r.top > 511) return;
        if (r.right > 1023) r.right = 1023;
        if (r.bottom > 511) r.bottom = 511;

        efx.dwSize = sizeof(efx);
        efx.dwFillColor = RGB32_16(c);
        IDirectDrawSurface_Blt(lpddDevice,
				   &r,
				   NULL,
				   NULL,
				   DDBLT_COLORFILL | DDBLT_WAIT,
				   &efx);
}

/**/
void sort_vertex(D3DTLVERTEX *v,char *order)
{
 int x,y;
 float m01,m02,divider;
 char c;

 for (x=2;x>0;x--)
	 for (y=0;y<x;y++)
		 if ((v+order[y])->sx > (v+order[y+1])->sx)
		 {
			 c = order[y];
			 order[y] = order[y+1];
			 order[y+1] = c;
		 }

 m01 = ((v+order[1])->sy - (v+order[0])->sy);
 divider = ((v+order[1])->sx - (v+order[0])->sx);
 if (divider != 0) m01 = m01 / divider;

 m02 = ((v+order[2])->sy - (v+order[0])->sy);
 divider = ((v+order[2])->sx - (v+order[0])->sx);
 if (divider != 0) m02 = m02 / divider;

 if (m01 > m02)
	 {
		 c = order[1];
		 order[1] = order[2];
		 order[2] = c;
	 }
}

static struct {
	int alpha;
	D3DBLEND src,dst;
} blend[4] = {
/* 50%+50% */
		{ 127<<24,D3DBLEND_SRCALPHA,D3DBLEND_INVSRCALPHA},
/* 100%+100% */
		{ 255<<24,D3DBLEND_ONE,D3DBLEND_ONE},
/* 100%-100% */
		{ 255<<24,D3DBLEND_SRCCOLOR,D3DBLEND_ONE},
/* 100%+25% */
		{  63<<24,D3DBLEND_SRCALPHA,D3DBLEND_ONE}
};

/*
static int setblend(int tpage)
{
	int mode = (tpage>>7)&3;
	_SetRenderState(D3DRENDERSTATE_SRCBLEND ,blend[mode].src);
	_SetRenderState(D3DRENDERSTATE_DESTBLEND,blend[mode].dst);
	return blend[mode].alpha;
}

#define	SETCOLOR(color,c,tpage) \
	if (c.code&1) color=RGB_MAKE(255,255,255); \
	else color = COLOR2D3DCOLOR(c); \
	if (c.code&2) { \
		color|=setblend(tpage); \
		_SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE); \
	} else { \
		_SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE,FALSE); \
	}
*/

static void DrawSprite(Sprt *prim,int w,int h)
{
    LPD3DTLVERTEX  lpVertex; 
    D3DCache	  *tex;
    UINT8          localbuf[GPUMAXEB];
    int            x,y,u,v,c;
    UINT8         *buf;
    RECT r1,r2;

    tex = get_texture(gpu.tpage,prim->v[0].pad);

    x = prim->v[0].x + OFFX;
    y = prim->v[0].y + OFFY;
    u = prim->v[0].u;
    v = prim->v[0].v;

    CLIPPER(x, FRAME_W, w, XMIN, XMAX, u)
    CLIPPER(y, FRAME_H, h, YMIN, YMAX, v)
/*
    r2.left = prim->v[0].x + OFFX;
    r2.right = r2.left + w;
    r2.top  = prim->v[0].y + OFFY;
    r2.bottom = r2.top + h;
    r1.left = prim->v[0].u;
    r1.right = r1.left + w;
    r1.top = prim->v[0].v;
    r1.bottom = r1.top + h;
*/
    r2.left = x;
    r2.right = r2.left + w;
    r2.top  = y;
    r2.bottom = r2.top + h;
    r1.left = u;
    r1.right = r1.left + w;
    r1.top = v;
    r1.bottom = r1.top + h;

    IDirectDrawSurface_Blt(lpddDevice,
			   &r2,
			   tex->surface,
			   &r1,
			   DDBLT_WAIT | DDBLT_KEYSRC,
			   NULL);

/*
    memset(localbuf,0,GPUMAXEB);

    c = RGB_MAKE(prim->c.r,prim->c.g,prim->c.b);

	lpVertex = (LPD3DTLVERTEX)localbuf;


	lpVertex[0].sx = lpVertex[2].sx = (D3DVALUE)(x);
	lpVertex[0].sy = lpVertex[1].sy = (D3DVALUE)(y);
	lpVertex[1].sx = lpVertex[3].sx = (D3DVALUE)(x+w);
	lpVertex[2].sy = lpVertex[3].sy = (D3DVALUE)(y+h);

	lpVertex[0].tu = lpVertex[2].tu = (D3DVALUE)(u)/255;
	lpVertex[0].tv = lpVertex[1].tv = (D3DVALUE)(v)/255;
	lpVertex[1].tu = lpVertex[3].tu = (D3DVALUE)(u+w)/255;
	lpVertex[2].tv = lpVertex[3].tv = (D3DVALUE)(v+h)/255;
	lpVertex += 4;

	for(x=0;x<4;x++) {
		lpVertex[x].sz = 0.0f;
		lpVertex[x].rhw = 1.0f;
		lpVertex[x].color = c;
		lpVertex[x].specular = RGB_MAKE(0,0,0);
	}

// render
	tex = get_texture(gpu.tpage,prim->v[0].pad);
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID, buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  FALSE,         buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_FLAT, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, tex->handle,   buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,TRUE,			buf);
	
	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 4, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(2, buf);

	((LPD3DTRIANGLE)buf)->wV1    = 0;
	((LPD3DTRIANGLE)buf)->wV2    = 1;
	((LPD3DTRIANGLE)buf)->wV3    = 2;
	((LPD3DTRIANGLE)buf)->wFlags = 0;
	buf = (void *)(((LPD3DTRIANGLE)(buf)) + 1);

	((LPD3DTRIANGLE)buf)->wV1    = 2;
	((LPD3DTRIANGLE)buf)->wV2    = 1;
	((LPD3DTRIANGLE)buf)->wV3    = 3;
	((LPD3DTRIANGLE)buf)->wFlags = 0;
	buf = (void *)(((LPD3DTRIANGLE)(buf)) + 1);

	OP_EXIT(buf);

	RenderScene(localbuf, 4, buf - (UINT8 *)lpVertex);
*/
}


static void DrawTile(Tile *prim,int w,int h)
{
    int x,y;

    x = prim->v[0].x + OFFX;
    y = prim->v[0].y + OFFY;

    if (x>XMAX || x+w<XMIN ||
        y>YMAX || y+h<YMIN ||
        w<=0   || h<=0) return;
    if (x<XMIN) x=XMIN;
    else if (x+w>XMAX) w=XMAX-x;
    if (y<YMIN) y=YMIN;
    else if (y+h>YMAX) h=YMAX-y;

	DrawBoxNoClip(x,y,w,h,
				  RGB_MAKE(prim->c.r,prim->c.g,prim->c.b));

}


void gpu_move(int x0,int y0,int x1,int y1,int w,int h)
{
        RECT    r1,r2;

        r1.left   = x0;
        r1.right  = x0+w;
        r1.top    = y0;
        r1.bottom = y0+h;
		r2.left   = x1;
        r2.right  = x1+w;
        r2.top    = y1;
        r2.bottom = y1+h;

        if (r1.left   > 1023) return;
        if (r1.top    > 511)  return;
        if (r2.left   > 1023) return;
        if (r2.top    > 511)  return;

        if (r1.right  > 1023) r1.right = 1023;
        if (r1.bottom > 511)  r1.bottom = 511;

        IDirectDrawSurface_Blt(lpddDevice,
							   &r2,
							   lpddDevice,
							   &r1,
							   DDBLT_WAIT,
							   NULL);
}

void gpu_polyF3(PolyF3 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	POINT_F       *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y,c;
	UINT8         *buf;
 
	p = prim->v;
	c = RGB_MAKE(prim->c.r,prim->c.g,prim->c.b);

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<3;x++)
	{
		order[x] = x;
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = c;
		lpVertex->specular = 0;
		lpVertex->tu = 0;
		lpVertex->tv = 0;
	    lpVertex++; p++;
	}
// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID, buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  FALSE,         buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_FLAT, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, (UINT32)NULL,  buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,FALSE,			buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 3, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(1, buf);

	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 3, buf - (UINT8 *)lpVertex);
}

void gpu_polyF4(PolyF4 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	POINT_F       *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y,c;
	UINT8         *buf;
 
	p = prim->v;
	c = RGB_MAKE(prim->c.r,prim->c.g,prim->c.b);

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<4;x++)
	{
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = c;
		lpVertex->specular = 0;
		lpVertex->tu = 0;
		lpVertex->tv = 0;
	    lpVertex++; p++;
	}
// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID, buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  FALSE,         buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_FLAT, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, (UINT32)NULL,  buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,FALSE,			buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 4, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(2, buf);

	order[0] = 0; order[1] = 1; order[2] = 2;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	order[0] = 1; order[1] = 2; order[2] = 3;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 4, buf - (UINT8 *)lpVertex);
}

void gpu_polyG3(PolyG3 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	POINT_G       *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y;
	UINT8         *buf;
 
	p = prim->v;

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<3;x++)
	{
		order[x] = x;
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = RGB_MAKE(p->c.r,p->c.g,p->c.b);
		lpVertex->specular = 0;
		lpVertex->tu = 0;
		lpVertex->tv = 0;
	    lpVertex++; p++;
	}

	sort_vertex((LPD3DTLVERTEX)localbuf,order);

// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID,    buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  TRUE,             buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_GOURAUD, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, (UINT32)NULL,     buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,FALSE,			buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 3, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(1, buf);
	FILL_TRIANGLE(0,buf);
	OP_EXIT(buf);

	RenderScene(localbuf,3,buf - (UINT8 *)lpVertex);
}

void gpu_polyG4(PolyG4 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	POINT_G       *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y;
	UINT8         *buf;
 
	p = prim->v;

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<4;x++)
	{
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = RGB_MAKE(p->c.r,p->c.g,p->c.b);
		lpVertex->specular = 0;
		lpVertex->tu = 0;
		lpVertex->tv = 0;
	    lpVertex++; p++;
	}

// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID,    buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  TRUE,             buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_GOURAUD, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, (UINT32)NULL,     buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,FALSE,			buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 4, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(2, buf);

	order[0] = 0; order[1] = 1; order[2] = 2;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	order[0] = 1; order[1] = 2; order[2] = 3;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 4, buf - (UINT8 *)lpVertex);
}

void gpu_polyFT3(PolyFT3 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	POINT_FT      *p;
	D3DCache	  *tex;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y,c;
	UINT8         *buf;
 
	p = prim->v;
	c = RGB_MAKE(prim->c.r,prim->c.g,prim->c.b);

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<3;x++)
	{
		order[x] = x;
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = c;
		lpVertex->specular = 0;
		lpVertex->tu = (D3DVALUE)(p->u)/255;
		lpVertex->tv = (D3DVALUE)(p->v)/255;
	    lpVertex++; p++;
	}
	tex = get_texture(prim->v[1].pad,prim->v[0].pad);
// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID, buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  FALSE,         buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_FLAT, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, tex->handle,   buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,TRUE,			buf);
//	STATE_DATA(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE, buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 3, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(1, buf);

	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 3, buf - (UINT8 *)lpVertex);
}

void gpu_polyFT4(PolyFT4 *prim)
{
    LPD3DTLVERTEX  lpVertex;
	D3DCache	  *tex;
	POINT_FT      *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y,c;
	UINT8         *buf;
 
	p = prim->v;
	c = RGB_MAKE(prim->c.r,prim->c.g,prim->c.b);

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<4;x++)
	{
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = c;
		lpVertex->specular = 0;
		lpVertex->tu = (D3DVALUE)(p->u)/255;
		lpVertex->tv = (D3DVALUE)(p->v)/255;
	    lpVertex++; p++;
	}
	tex = get_texture(prim->v[1].pad,prim->v[0].pad);

// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID, buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  FALSE,         buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_FLAT, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, tex->handle,   buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,TRUE,			buf);
//	STATE_DATA(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE, buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 4, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(2, buf);

	order[0] = 0; order[1] = 1; order[2] = 2;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	order[0] = 1; order[1] = 2; order[2] = 3;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 4, buf - (UINT8 *)lpVertex);
}

void gpu_polyGT3(PolyGT3 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	D3DCache	  *tex;
	POINT_GT      *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y;
	UINT8         *buf;
 
	p = prim->v;

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<3;x++)
	{
		order[x] = x;
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = RGB_MAKE(p->c.r,p->c.g,p->c.b);
		lpVertex->specular = 0;
		lpVertex->tu = (D3DVALUE)(p->u)/255;
		lpVertex->tv = (D3DVALUE)(p->v)/255;
	    lpVertex++; p++;
	}

	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	tex = get_texture(prim->v[1].pad,prim->v[0].pad);

// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID,    buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  TRUE,             buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_GOURAUD, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, tex->handle,      buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,TRUE,			buf);
//	STATE_DATA(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE, buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 3, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(1, buf);
	FILL_TRIANGLE(0,buf);
	OP_EXIT(buf);

	RenderScene(localbuf,3,buf - (UINT8 *)lpVertex);
}

void gpu_polyGT4(PolyGT4 *prim)
{
    LPD3DTLVERTEX  lpVertex; 
	D3DCache	  *tex;
	POINT_GT      *p;
	UINT8          localbuf[GPUMAXEB];
	UINT8          order[4];
	int            x,y;
	UINT8         *buf;
 
	p = prim->v;

	lpVertex = (LPD3DTLVERTEX)localbuf;

	for (x=0;x<4;x++)
	{
		y = p->x+OFFX;
		if (y<XMIN) y = XMIN;
		else 
		if (y>XMAX) y = XMAX;
		lpVertex->sx = (D3DVALUE)y;
		y = p->y+OFFY;
		if (y < YMIN) y = YMIN;
		else 
		if (y>YMAX) y = YMAX;
		lpVertex->sy = (D3DVALUE)y;
		lpVertex->sz = 0;
		lpVertex->rhw = 1;
		lpVertex->color = RGB_MAKE(p->c.r,p->c.g,p->c.b);
		lpVertex->specular = 0;
		lpVertex->tu = (D3DVALUE)(p->u)/255;
		lpVertex->tv = (D3DVALUE)(p->v)/255;
	    lpVertex++; p++;
	}
	tex = get_texture(prim->v[1].pad,prim->v[0].pad);

// render
	buf = (UINT8 *)lpVertex;
	OP_STATE_RENDER(5,buf);
	STATE_DATA(D3DRENDERSTATE_FILLMODE,      D3DFILL_SOLID,    buf);
	STATE_DATA(D3DRENDERSTATE_DITHERENABLE,  TRUE,             buf);
	STATE_DATA(D3DRENDERSTATE_SHADEMODE,     D3DSHADE_GOURAUD, buf);
    STATE_DATA(D3DRENDERSTATE_TEXTUREHANDLE, tex->handle,      buf);
	STATE_DATA(D3DRENDERSTATE_COLORKEYENABLE,TRUE,			buf);
//	STATE_DATA(D3DRENDERSTATE_ALPHABLENDENABLE,TRUE, buf);

	OP_PROCESS_VERTICES(1, buf);
	PROCESSVERTICES_DATA(D3DPROCESSVERTICES_COPY, 0, 4, buf);

	if( QWORD_ALIGNED( buf ) ) { OP_NOP( buf ); }

	OP_TRIANGLE_LIST(2, buf);

	order[0] = 0; order[1] = 1; order[2] = 2;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	order[0] = 1; order[1] = 2; order[2] = 3;
	sort_vertex((LPD3DTLVERTEX)localbuf,order);
	FILL_TRIANGLE(0,buf);

	OP_EXIT(buf);

	RenderScene(localbuf, 4, buf - (UINT8 *)lpVertex);
}

void gpu_lineF2(LineF2 *prim)
{
}

void gpu_lineF3(LineF3 *prim)
{
}

void gpu_lineF4(LineF4 *prim)
{
}

void gpu_lineG2(LineG2 *prim)
{
}

void gpu_lineG3(LineG3 *prim)
{
}

void gpu_lineG4(LineG4 *prim)
{
}

void gpu_fill(Fill *prim)
{
	DrawBoxNoClip(prim->v[0].x,prim->v[0].y,
				  prim->w,prim->h,
				  RGB_MAKE(prim->c.r,prim->c.g,prim->c.b));
}

void gpu_tile(Tile *prim)
{
    DrawTile(prim,prim->w,prim->h);
}

void gpu_tile8(Tile8 *prim)
{
    DrawTile((Tile *)prim,8,8);
}

void gpu_tile16(Tile16 *prim)
{
    DrawTile((Tile *)prim,16,16);
}

void gpu_tile1(Tile1 *prim)
{
    DrawTile((Tile *)prim,1,1);
}

void gpu_sprt(Sprt *prim)
{
	DrawSprite(prim,prim->w,prim->h);
}

void gpu_sprt8(Sprt *prim)
{
	DrawSprite(prim,8,8);
}

void gpu_sprt16(Sprt *prim)
{
	DrawSprite(prim,16,16);
}
