/*
#ifdef _WIN32
#define WIN32 _WIN32
#endif
*/
#include <stdio.h>
#include <memory.h>
#include "type.h"
#include "audio.h"
#include "sound.h"

#define	MAXVOICE	24
#define	WCACHE_MAX	24

static HAC hVoice[MAXVOICE];
static AUDIOWAVE wave[MAXVOICE];

int sound_init(void)
{
	AUDIOINFO info;
	int i;

	if (AInitialize()) return -1;
	info.nDeviceId = AUDIO_DEVICE_MAPPER;
	info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;
	info.nSampleRate = 44100;
	if (AOpenAudio(&info)) return -1;

	if (AOpenVoices(MAXVOICE)) {
		ACloseAudio();
		return -1;
	}

	for(i=0;i<MAXVOICE;i++) {
		ACreateAudioVoice(&hVoice[i]);
	}

	return 0;
}

int sound_term(void)
{
	int i;

	for(i=0;i<MAXVOICE;i++) {
		AStopVoice(hVoice[i]);
		ADestroyAudioVoice(hVoice[i]);
	}
	ACloseVoices();
	ACloseAudio();
	return 0;
}

static void err(char *func,int errcode)
{
	char buf[256];
	AGetErrorText(errcode,buf,sizeof(buf));
	printf("%s:%s\n",func,buf);
}

static int master_vol_l,master_vol_r;
void SetMainVolume(int vol_l,int vol_r)
{
	master_vol_l = vol_l;
	master_vol_r = vol_r;
}

static void create_wave(AUDIOWAVE *w,short *data,int size,int loop_start)
{
	int ret;
	if (w->dwHandle) {
		ret = ADestroyAudioData(w);
		if (ret) err("ADestroyAudioData",ret);
	}
	w->lpData = 0;
	w->dwHandle = 0;

	w->dwLength = size*2;
	w->nSampleRate = 44100;
	w->wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_MONO;
	if (loop_start>=0) {
		w->wFormat|= AUDIO_FORMAT_LOOP;
		w->dwLoopStart = loop_start*2;
		w->dwLoopEnd = size*2;
	} else {
		w->dwLoopStart = 0;
		w->dwLoopEnd = 0;
	}
	ret = ACreateAudioData(w);
	if (ret) err("ACreateAudioData",ret);
//	printf("handle %d ptr:%p\n",w->dwHandle,w->lpData);
	memcpy(w->lpData,data,w->dwLength);
	ret = AWriteAudioData(w,0,w->dwLength);
	if (ret) err("AWriteAudioData",ret);
}

void wav_stop(int ch)
{
	AStopVoice(hVoice[ch]);
}

void wav_play(int ch,short *data,int size,int freq,int vol_l,int vol_r,int loop_start)
{
	HAC voice;
	int vol,pan;
	AUDIOWAVE *w;

	int ret;

	wav_stop(ch);

	w = &wave[ch];
	create_wave(w,data,size,loop_start);

	vol_l = (vol_l*master_vol_l)>>14;
	vol_r = (vol_l*master_vol_r)>>14;

	/* vol_l,vol_r -> vol,pan */
	vol = vol_l;
	if (vol<vol_r) vol = vol_r;
	if (vol>0) {
		if (vol_l>vol_r)
			pan = vol_r*128/vol_l;
		else
			pan = 255 - vol_l*128/vol_r;
	} else {
		pan=128;
	}
	/* psx vol:0..3fff -> seal vol:0..64 */
	vol>>=8;

	voice = hVoice[ch];
	ret = APrimeVoice(voice,w);
	if (ret) err("APrimeVoice",ret);
	ASetVoiceFrequency(voice,freq);
	if (ret) err("ASetVoiceFrequency",ret);
	ret = ASetVoiceVolume(voice,vol);
	if (ret) err("ASetVoiceVolume",ret);
	ret = ASetVoicePanning(voice,pan);
	if (ret) err("ASetVoicePanning",ret);
	ret = AStartVoice(voice);
}

int wav_status(int ch)
{
	BOOL status;
	AGetVoiceStatus(hVoice[ch],&status);
	return status;
}
